/* FreeS/WAN config file parser (confread.c)
 * Copyright (C) 2001-2002 Mathieu Lafon - Arkoon Network Security
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * RCSID $Id: confread.c,v 1.5 2004/01/20 20:47:42 mcr Exp $
 */

#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>

#include "parser.h"
#include "confread.h"
#include "interfaces.h"
#include "starterlog.h"

static char _tmp_err[512];

#define POLICY_ONLY_CONN(conn) if(conn->options[KBF_AUTO] == STARTUP_START) { conn->options[KBF_AUTO]=STARTUP_ROUTE; }

static void default_values (struct starter_config *cfg)
{
	if (!cfg) return;
	memset(cfg, 0, sizeof(struct starter_config));

	cfg->setup.options[KBF_FRAGICMP] = TRUE;
	cfg->setup.options[KBF_HIDETOS]  = TRUE;
	cfg->setup.options[KBF_UNIQUEIDS]= FALSE;
	cfg->setup.options[KBF_TYPE] = KS_TUNNEL;

	cfg->conn_default.policy = POLICY_ENCRYPT | POLICY_TUNNEL | POLICY_PSK;

	cfg->conn_default.options[KBF_IKELIFETIME] = OAKLEY_ISAKMP_SA_LIFETIME_DEFAULT;
	cfg->conn_default.options[KBF_SALIFETIME]  = SA_LIFE_DURATION_DEFAULT;
	cfg->conn_default.options[KBF_REKEYMARGIN] = SA_REPLACEMENT_MARGIN_DEFAULT;
	cfg->conn_default.options[KBF_REKEYFUZZ]   = SA_REPLACEMENT_FUZZ_DEFAULT;
	cfg->conn_default.options[KBF_KEYINGTRIES] = SA_REPLACEMENT_RETRIES_DEFAULT;

	cfg->conn_default.left.addr_family = AF_INET;
	anyaddr(AF_INET, &cfg->conn_default.left.addr);
	anyaddr(AF_INET, &cfg->conn_default.left.nexthop);
	cfg->conn_default.right.addr_family = AF_INET;
	anyaddr(AF_INET, &cfg->conn_default.right.addr);
	anyaddr(AF_INET, &cfg->conn_default.right.nexthop);

	cfg->conn_default.options[KBF_AUTO] = STARTUP_NO;
	cfg->conn_default.state = STATE_LOADED;
}

#define ERR_FOUND(args...) \
	{ if (perr && (*perr==NULL)) { \
		snprintf(_tmp_err, sizeof(_tmp_err)-1, ## args); \
		*perr = xstrdup(_tmp_err); } \
	err++; }

#define KW_POLICY_FLAG(val,fl) if(conn->options[val]) \
	{ \
	    conn->policy |= fl; \
	} else { \
	    conn->policy &= ~fl;\
	}


static void free_list(char **list)
{
	char **s;
	for (s=list; *s; s++) free(*s);
	free(list);
}

static char **new_list(char *value)
{
	char *val, *b, *e, *end, **ret;
	int count;
	val = value ? xstrdup(value) : NULL;
	if (!val) return NULL;
	end = val + strlen(val);
	for (b=val, count=0; b<end; ) {
		for (e=b; ((*e!=' ') && (*e!='\0')); e++);
		*e = '\0';
		if (e!=b) { count++; }
		b=e+1;
	}
	if (count==0) {
		free(val);
		return NULL;
	}
	ret = (char **)malloc((count+1) * sizeof(char *));
	if (!ret) {
		free(val);
		return NULL;
	}
	for (b=val, count=0; b<end; ) {
		for (e=b; (*e!='\0'); e++);
		if (e!=b) {
			ret[count++] = xstrdup(b);
		}
		b=e+1;
	}
	ret[count] = NULL;
	free(val);
	return ret;
}

static int load_setup (struct starter_config *cfg, struct config_parsed *cfgp,
	char **perr)
{
	unsigned int err = 0;
	struct kw_list *kw;

	for (kw=cfgp->config_setup; kw; kw=kw->next) {

	    /*
	     * the parser already made sure that only config keywords were used,
	     * but we double check!
	     */
	    assert(kw->keyword.keydef->validity & kv_config);

	    switch(kw->keyword.keydef->type)
	    {
	    case kt_string:
	    case kt_filename:
	    case kt_dirname:
	    case kt_loose_enum:
		/* all treated as strings for now */
		assert(kw->keyword.keydef->field < sizeof(cfg->setup.strings));
		if(cfg->setup.strings[kw->keyword.keydef->field]) free(cfg->setup.strings[kw->keyword.keydef->field]);
		cfg->setup.strings[kw->keyword.keydef->field] = xstrdup(kw->string);
		break;

	    case kt_appendstring:
		assert(kw->keyword.keydef->field < KEY_STRINGS_MAX);
		if(!cfg->setup.strings[kw->keyword.keydef->field])
		{
		    cfg->setup.strings[kw->keyword.keydef->field] = xstrdup(kw->string);
		} else {
		    int len;
		    char *s;
		    
		    len = strlen(cfg->setup.strings[kw->keyword.keydef->field])+1;
		    len += strlen(kw->string)+1;
		    
		    /* allocate the string */
		    s = cfg->setup.strings[kw->keyword.keydef->field];
		    s = xrealloc(s, len);
		    strncat(s, " ", len);
		    strncat(s, kw->string, len);
		    
		    cfg->setup.strings[kw->keyword.keydef->field] = s;
		}
		break;
		
	    case kt_list:
	    case kt_bool:
	    case kt_invertbool:
	    case kt_enum:
	    case kt_number:
	    case kt_time:
	    case kt_percent:
		/* all treated as a number for now */
		assert(kw->keyword.keydef->field < sizeof(cfg->setup.options));
		cfg->setup.options[kw->keyword.keydef->field] = kw->number;
		break;

	    case kt_bitstring:
	    case kt_rsakey:
	    case kt_ipaddr:
	    case kt_subnet:
	    case kt_idtype:
		err++;
		break;
	    }
	}
		
	/* now process some things with specific values */
	
	/* interfaces has to be chopped up */
	if (cfg->setup.interfaces) free_list(cfg->setup.interfaces);
	cfg->setup.interfaces = new_list(cfg->setup.strings[KSF_INTERFACES]);

	return err;
}

static int validate_end(struct starter_conn *conn_st
			, struct starter_end *end
			, bool left, char **perr)
{
    err_t er = NULL;
    int err=0;
    
    end->addrtype=end->options[KNCF_IP];

    /* validate the KSCF_IP/KNCF_IP */
    switch(end->addrtype)
    {
    case KH_ANY:
	anyaddr(AF_INET, &(end->addr));
	break;

    case KH_IFACE:
	assert(end->strings[KSCF_IP] != NULL);

	if (end->iface) free(end->iface);
	end->iface = xstrdup(end->strings[KNCF_IP]);
	if (starter_iface_find(end->iface, AF_INET, &(end->addr),
			       &(end->nexthop)) == -1) {
	    conn_st->state = STATE_INVALID;
	}
	break;
	
    case KH_IPADDR:
	assert(end->strings[KSCF_IP] != NULL);

	er = ttoaddr(end->strings[KNCF_IP], 0, AF_INET, &(end->addr));
	if (er) ERR_FOUND("bad addr %s=%s [%s]", (left ? "left" : "right"), end->strings[KNCF_IP], er);
	break;
	
    case KH_DEFAULTROUTE:
    case KH_OPPO:
    case KH_OPPOGROUP:
    case KH_GROUP:
	break;

    case KH_NOTSET:
	break;
    }

    /* validate the KSCF_SUBNET */
    if(end->strings[KSCF_SUBNET] != NULL)
    {
	char *value = end->strings[KSCF_SUBNET];

#ifdef VIRTUAL_IP
        if ( ((strlen(value)>=6) && (strncmp(value,"vhost:",6)==0)) ||
	     ((strlen(value)>=5) && (strncmp(value,"vnet:",5)==0)) ) {
	    er = NULL;
	    end->virt = xstrdup(value);
	    if (!end->virt) ERR_FOUND("can't alloc memory");
	}
	else {
	    end->has_client = TRUE;
	    er = ttosubnet(value, 0, AF_INET, &(end->subnet));
	}
#else
	end->has_client = TRUE;
#ifdef X509
	end->has_client_wildcard = FALSE;
#endif
	er = ttosubnet(value, 0, AF_INET, &(end->subnet));
#endif
	if (er) ERR_FOUND("bad subnet %s=%s [%s]", (left ? "leftsubnet" : "rightsubnet"), value, er);
    }

    /* validate the KSCF_NEXTHOP */
    if(end->strings[KSCF_NEXTHOP] != NULL)
    {
	char *value = end->strings[KSCF_NEXTHOP];
	
	er = ttoaddr(value, 0, AF_INET, &(end->nexthop));
	if (er) ERR_FOUND("bad addr %s=%s [%s]", (left ? "lextnexthop" : "rightnexthop"), value, er);
    }

    /* validate the KSCF_ID */
    if(end->strings[KSCF_ID] != NULL)
    {
	char *value = end->strings[KSCF_ID];
	
	if (end->id) free(end->id);
	end->id = xstrdup(value);
    }

    /* validate the KSCF_RSAKEY1/RSAKEY2 */
    if(end->strings[KSCF_RSAKEY1] != NULL)
    {
	char *value = end->strings[KSCF_RSAKEY1];

	if (end->rsakey1) free(end->rsakey1);
	end->rsakey1 = xstrdup(value);
    }
    if(end->strings[KSCF_RSAKEY2] != NULL)
    {
	char *value = end->strings[KSCF_RSAKEY2];

	if (end->rsakey2) free(end->rsakey2);
	end->rsakey2 = xstrdup(value);
    }

    return err;
}


static int load_conn (struct starter_conn *conn
		      , struct config_parsed *cfgp
		      , struct kw_list *kw
		      , char **perr)
{
    volatile unsigned int err;
    ksf    *the_strings;
    knf    *the_options;
    volatile int i;
	
    err = 0;
    i = 0;

	for ( ; kw; kw=kw->next)
	{
	    i++;
	    the_strings = &conn->strings;
	    the_options = &conn->options;

	    if((kw->keyword.keydef->validity & kv_conn) == 0)
	    {
		/* this isn't valid in a conn! */
		starter_log(LOG_LEVEL_INFO,
			    "keyword '%s' is not valid in a conn (%s)\n",
			    kw->keyword.keydef->keyname, conn->name);
		continue;
	    }

	    if(kw->keyword.keydef->validity & kv_leftright)
	    {
		if(kw->keyword.keyleft)
		{
		    the_strings = &conn->left.strings;
		    the_options = &conn->left.options;
		} else {
		    the_strings = &conn->right.strings;
		    the_options = &conn->right.options;
		}
	    }

	    assert(kw->keyword.keydef != NULL);
	    switch(kw->keyword.keydef->type)
	    {
	    case kt_string:
	    case kt_filename:
	    case kt_dirname:
	    case kt_bitstring:
	    case kt_ipaddr:
	    case kt_subnet:
	    case kt_idtype:
		/* all treated as strings for now */
		assert(kw->keyword.keydef->field < KEY_STRINGS_MAX);
		if((*the_strings)[kw->keyword.keydef->field])
		{
		    free((*the_strings)[kw->keyword.keydef->field]);
		}
		(*the_strings)[kw->keyword.keydef->field] = xstrdup(kw->string);
		break;

	    case kt_appendstring:
		assert(kw->keyword.keydef->field < KEY_STRINGS_MAX);
		if(!(*the_strings)[kw->keyword.keydef->field])
		{
		    (*the_strings)[kw->keyword.keydef->field] = xstrdup(kw->string);
		} else {
		    int len;
		    char *s;
		    
		    len = strlen((*the_strings)[kw->keyword.keydef->field])+1;
		    len += strlen(kw->string)+1;
		    
		    /* allocate the string */
		    s = (*the_strings)[kw->keyword.keydef->field];
		    s = xrealloc(s, len);
		    strncat(s, " ", len);
		    strncat(s, kw->string, len);
		    
		    (*the_strings)[kw->keyword.keydef->field] = s;
		}
		break;
		
	    case kt_rsakey:
	    case kt_loose_enum:
		assert(kw->keyword.keydef->field < KEY_STRINGS_MAX);
		assert(kw->keyword.keydef->field < KEY_NUMERIC_MAX);

		(*the_options)[kw->keyword.keydef->field] = kw->number;
		if(kw->number == LOOSE_ENUM_OTHER)
		{
		    assert(kw->keyword.string != NULL);
		    if((*the_strings)[kw->keyword.keydef->field]) free((*the_strings)[kw->keyword.keydef->field]);
		    (*the_strings)[kw->keyword.keydef->field] = xstrdup(kw->keyword.string);
		} 
		break;

	    case kt_list:
	    case kt_bool:
	    case kt_invertbool:
	    case kt_enum:
	    case kt_number:
	    case kt_time:
	    case kt_percent:
		/* all treated as a number for now */
		assert(kw->keyword.keydef->field < KEY_NUMERIC_MAX);
		(*the_options)[kw->keyword.keydef->field] = kw->number;
		break;
	    }
	}

	/* validate some things */

	/* preview of getting rid of transport mode */
	if(conn->options[KBF_TYPE] == KS_TRANSPORT)
	{
	    extern struct keyword_enum_values kw_type_list;

	    starter_log(LOG_LEVEL_INFO, "conn %s - only tunnel mode is supported, not %s"
			, conn->name
			, keyword_name(&kw_type_list, conn->options[KBF_TYPE]));
	    
	    POLICY_ONLY_CONN(conn);
	}

	KW_POLICY_FLAG(KBF_TYPE, POLICY_TUNNEL);
	KW_POLICY_FLAG(KBF_COMPRESS, POLICY_COMPRESS);
	KW_POLICY_FLAG(KBF_PFS,  POLICY_PFS);
	
	/* reset authby flags */
	conn->policy &= ~(POLICY_ID_AUTH_MASK);
	conn->policy |= conn->options[KBF_AUTHBY];

	KW_POLICY_FLAG(KBF_REKEY, POLICY_DONT_REKEY);
	KW_POLICY_FLAG(KBF_COMPRESS, POLICY_COMPRESS);

	err += validate_end(conn, &conn->left,  TRUE, perr);
	err += validate_end(conn, &conn->right, FALSE,perr);

	return err;
}


static void conn_default (char *name, struct starter_conn *conn,
	struct starter_conn *def)
{
    int i;

	memcpy(conn, def, sizeof(struct starter_conn));

	conn->name = xstrdup(name);
#define CONN_STR(v) if (v) v=xstrdup(v)
	CONN_STR(conn->left.iface);
	CONN_STR(conn->left.id);
	CONN_STR(conn->left.rsakey1);
	CONN_STR(conn->left.rsakey2);
	CONN_STR(conn->right.iface);
	CONN_STR(conn->right.id);
	CONN_STR(conn->right.rsakey1);
	CONN_STR(conn->right.rsakey2);

	for(i=0; i<KSCF_MAX; i++)
	{
	    CONN_STR(conn->left.strings[i]);
	    CONN_STR(conn->right.strings[i]);
	}
	for(i=0 ;i<KSF_MAX; i++)
	{
	    CONN_STR(conn->strings[i]);
	}

#ifdef ALG_PATCH
	CONN_STR(conn->esp);
	CONN_STR(conn->ike);
#endif
#undef CONN_STR
}

struct starter_config *confread_load(const char *file, char **perr)
{
	struct starter_config *cfg = NULL;
	struct config_parsed *cfgp;
	struct section_list *sconn;
	struct starter_conn *conn;
	unsigned int err = 0;

	/**
	 * Load file
	 */
	cfgp = parser_load_conf(file, perr);
	if (!cfgp) return NULL;

	cfg = (struct starter_config *)malloc(sizeof(struct starter_config));
	if (!cfg) {
		if (perr) *perr = xstrdup("can't allocate mem in confread_load()");
		parser_free_conf(cfgp);
		return NULL;
	}

	/**
	 * Set default values
	 */
	default_values(cfg);

	/**
	 * Load setup
	 */
	err += load_setup(cfg, cfgp, perr);

	if(err) {return NULL;}

	/**
	 * Find %default conn
	 *
	 */
	for (sconn=cfgp->conn_first; (!err) && (sconn); sconn=sconn->next) {
		if (strcmp(sconn->name,"%default")==0) {
			starter_log(LOG_LEVEL_DEBUG, "Loading default conn");
			err += load_conn (&cfg->conn_default, cfgp, sconn->kw, perr);
		}
	}

	/**
	 * Load other conns
	 */
	for (sconn=cfgp->conn_first; (!err) && (sconn); sconn=sconn->next) {
		if (strcmp(sconn->name,"%default")==0) continue;
		starter_log(LOG_LEVEL_DEBUG, "Loading conn %s", sconn->name);
		conn = (struct starter_conn *)malloc(sizeof(struct starter_conn));
		memset(conn, 0, sizeof(struct starter_conn));
		if (!conn) {
			if (perr) *perr = xstrdup("can't allocate mem in confread_load()");
			parser_free_conf(cfgp);
			confread_free(cfg);
			return NULL;
		}
		conn_default(sconn->name, conn, &cfg->conn_default);
		conn->state = STATE_TO_ADD;
		conn->next = NULL;
		if (cfg->conn_last) cfg->conn_last->next = conn;
		cfg->conn_last = conn;
		if (!cfg->conn_first) cfg->conn_first = conn;
		err += load_conn (conn, cfgp, sconn->kw, perr);
	}

	parser_free_conf(cfgp);

	if (err) {
		confread_free(cfg);
		cfg = NULL;
	}

	return cfg;
}

#define FREE_STR(v) { if (v) { free(v); v=NULL; } }
#define FREE_LST(v) { if (v) { free_list(v); v=NULL; } }
static void confread_free_conn(struct starter_conn *conn)
{
    int i;
	FREE_STR(conn->left.iface);
	FREE_STR(conn->left.id);
	FREE_STR(conn->left.rsakey1);
	FREE_STR(conn->left.rsakey2);
	FREE_STR(conn->right.iface);
	FREE_STR(conn->right.id);
	FREE_STR(conn->right.rsakey1);
	FREE_STR(conn->right.rsakey2);
	for(i=0; i<KSCF_MAX; i++)
	{
	    FREE_STR(conn->left.strings[i]);
	    FREE_STR(conn->right.strings[i]);
	}
	for(i=0 ;i<KSF_MAX; i++)
	{
	    FREE_STR(conn->strings[i]);
	}

#ifdef ALG_PATCH
	FREE_STR(conn->esp);
	FREE_STR(conn->ike);
#endif
#ifdef VIRTUAL_IP
	FREE_STR(conn->left.virt);
	FREE_STR(conn->right.virt);
#endif
}

void confread_free(struct starter_config *cfg)
{
    int i;
	struct starter_conn *conn, *c;
	FREE_LST(cfg->setup.interfaces);
#ifdef VIRTUAL_IP
	FREE_STR(cfg->setup.virtual_private);
#endif
	for(i=0 ;i<KSF_MAX; i++)
	{
	    FREE_STR(cfg->setup.strings[i]);
	}
	confread_free_conn(&(cfg->conn_default));
	for (conn=cfg->conn_first; conn; ) {
		c = conn;
		conn = conn->next;
		confread_free_conn(c);
		free(c);
	}
	free(cfg);
}
#undef FREE_STR

/*
 * Local Variables:
 * c-basic-offset:4
 * c-style: pluto
 * End:
 */
